#include "server_env.hpp"
#include "utils/envmanager.hpp"
#include "utils/directories.hpp"
#include "utils/utils.hpp"
#include "services.hpp"

#include <iostream>
#include <memory>
#include <filesystem>
namespace dropshell {

bool server_env::is_valid() const {
    return mValid;
}

server_env::server_env(const std::string& server_name) : mValid(false), mServer_name(server_name) {
    if (server_name.empty())
    {
        std::cerr << "Warning: Server name is empty, passed to server_env constructor." << std::endl;
        return;
    }
        
    // Construct the full path to server.env
    std::string env_path = get_local_server_env_path(server_name);
    
    // Check if file exists
    if (!std::filesystem::exists(env_path)) {
        std::cerr << "Server environment file not found: " + env_path << std::endl;
        return;
    }

    try {
        // Use envmanager to handle the environment file
        m_env_manager = std::unique_ptr<envmanager>(new envmanager(env_path));
        m_env_manager->load();
        
        // Get all variables
        m_env_manager->get_all_variables_substituted(variables);

        // Verify required variables exist
        for (const auto& var : {"SSH_HOST", "SSH_USER", "SSH_PORT", "DROPSHELL_DIR"}) {
            if (variables.find(var) == variables.end()) {
                // Print the variables identified in the file
                std::cout << "Variables identified in the file:" << std::endl;
                for (const auto& v : variables) {
                    std::cout << "    " << v.first << std::endl;
                }
                throw std::runtime_error("Missing required variable: " + std::string(var));
            }
        }
        mValid = true;

    } catch (const std::exception& e) {
        std::cerr << "Failed to parse server environment file: " + std::string(e.what()) << std::endl;
    }    
}

std::string server_env::get_variable(const std::string& name) const {
    if (!m_env_manager) {
        return "";
    }
    return m_env_manager->get_variable_substituted(name);
}

const std::map<std::string, std::string> &server_env::get_variables() const
{
    return variables;
}

std::string server_env::get_SSH_HOST() const {
    return get_variable("SSH_HOST");
}

std::string server_env::get_SSH_USER() const {
    return get_variable("SSH_USER");
}

std::string server_env::get_SSH_PORT() const {
    return get_variable("SSH_PORT");
}

std::string server_env::get_DROPSHELL_DIR() const {
    return get_variable("DROPSHELL_DIR");
}

// Helper method implementations
std::string server_env::construct_ssh_cmd() const {
    std::stringstream ssh_cmd;
    ssh_cmd << "ssh -p " << get_SSH_PORT() << " "
            << get_SSH_USER() << "@" << get_SSH_HOST() << " ";
    return ssh_cmd.str();
}

std::string server_env::construct_standard_command_run_cmd(const std::string &service_name, const std::string &command) const
{
    std::string remote_service_template_path = get_remote_service_template_path(mServer_name,service_name);
    std::string remote_service_config_path = get_remote_service_config_path(mServer_name,service_name);

    std::string script_path = remote_service_template_path + "/" + command + ".sh";
    std::string run_cmd = "'cd " + quote(remote_service_template_path) + 
                        " && /bin/bash "+quote(script_path)+" "+quote(remote_service_config_path)+"'";
    return run_cmd;
}

bool server_env::check_remote_dir_exists(const std::string &dir_path) const
{
    std::string check_dir_cmd = construct_ssh_cmd() + "'test -d " + quote(dir_path) + "'";
    if (system(check_dir_cmd.c_str()) != 0) {
        std::cerr << "Error: Directory not found on remote server:" << std::filesystem::path(dir_path).filename().string() << std::endl;
        return false;
    }
    return true;
}

bool server_env::check_remote_file_exists(const std::string& file_path) const {
    std::string check_cmd = construct_ssh_cmd() + "'test -f " + quote(file_path) + "'";
    if (system(check_cmd.c_str()) != 0) {
        std::cerr << "Error: File not found on remote server: " << std::filesystem::path(file_path).filename().string() << std::endl;
        return false;
    }
    return true;
}

bool server_env::check_remote_items_exist(const std::vector<std::string> &file_paths) const
{
    // convert file_paths to a single string, separated by spaces
    std::string file_paths_str;
    std::string file_names_str;
    for (const auto& file_path : file_paths) {
        file_paths_str += quote(file_path) + " ";
        file_names_str += std::filesystem::path(file_path).filename().string() + " ";
    }
    // check if all items in the vector exist on the remote server, in a single command.
    std::string check_cmd = construct_ssh_cmd() + "'for item in " + file_paths_str + "; do test -f $item; done'";
    if (system(check_cmd.c_str()) != 0) {
        std::cerr << "Error: Required items not found on remote server: " << file_names_str << std::endl;
        return false;
    }
    return true;
}

bool server_env::execute_ssh_command(const std::string& command, const std::string& error_msg) const {
    std::string full_cmd = construct_ssh_cmd() + command;
    return execute_local_command(full_cmd, error_msg);
}

bool server_env::execute_local_command(const std::string& command, const std::string& error_msg) {
    bool okay = (system(command.c_str()) == 0);

    if (!okay && !error_msg.empty())
        std::cerr << "Error: " << error_msg << std::endl;
    return okay;
}

bool server_env::execute_local_command_and_capture_output(const std::string &command, std::string &output)
{
    std::string full_cmd = command + " 2>&1";
    FILE *pipe = popen(full_cmd.c_str(), "r");
    if (!pipe) {
        std::cerr << "Error: Failed to execute command: " << command << std::endl;
        return false;
    }
    char buffer[128];
    while (fgets(buffer, sizeof(buffer), pipe) != nullptr) {
        output += buffer;
    }
    pclose(pipe);
    return true;
}


} // namespace dropshell 